# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Tests for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import hope
import itertools, pytest

from test.utilities import random, check, make_test, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype", dtypes)
def test_assignment(dtype):
    def fkt(a, b): a[:] = b
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [10])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_index_1d(dtype):
    def fkt(a, b): a[5] = b[3]
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [10])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_index_2d(dtype):
    def fkt(a, b): a[4, 2] = b[3, 4]
    (ao, ah), (b, _) = random(dtype, [5, 5]), random(dtype, [5, 5])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_index_2d_2(dtype):
    def fkt(a): a[1:4, 1:4] = 1
    (ao, ah) = random(dtype, [5, 5])
    
    fkt(ao), hope.jit(fkt)(ah)
    assert check(ao, ah)
    fkt(ao), hope.jit(fkt)(ah)
    assert check(ao, ah)
    
@pytest.mark.parametrize("dtype", dtypes)
def test_scalar(dtype):
    def fkt(a, b): a[3] = b
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_slice_1d_1(dtype):
    def fkt(a, b): a[2:5] = b[6:9]
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [10])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_slice_1d_2(dtype):
    def fkt(a, b): a[:5] = b[3:8]
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [10])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype", dtypes)
def test_slice_1d_3(dtype):
    def fkt(a, b): a[2:] = b[1:9]
    (ao, ah), (b, _) = random(dtype, [10]), random(dtype, [10])
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)
    fkt(ao, b), hope.jit(fkt)(ah, b)
    assert check(ao, ah)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_assign(a, b, c): c[:] = a


@pytest.mark.parametrize("dtype", dtypes)
def test_merged_slice(dtype):
    def fkt(a, b, c):
        for i in range(10):
            c[:, i] = a[i, :]
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, [10, 10]), random(dtype, [10, 10]), random(dtype, [10, 10])
    ao, ah, bo, bh = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype), (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype", dtypes)
def test_merged_slice_2(dtype):
    def fkt(a, c):
        a[1:] = c[:-1] + c[1:]
        c[:] = a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, [5]), random(dtype, [5])
    ao, ah = (ao / 4.).astype(dtype), (ah / 4.).astype(dtype)
    co, ch = (co / 4.).astype(dtype), (ch / 4.).astype(dtype)
    fkt(ao, co), hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co), hfkt(ah, ch)
    assert check(co, ch)


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_binary_mod(dtype, shape):
    if JENKINS and dtype == np.int8:
        pytest.skip("Fails on debian: dtype={0!s}".format(dtype))
    def fkt(a, b, c):
        c[:] = a % b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    if np.count_nonzero(bo == 0) > 0: bo[bo == 0] += 1
    if np.count_nonzero(bh == 0) > 0: bh[bh == 0] += 1
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_binary_lshifts(dtype, shape):
    def fkt(a, b, c):
        c[:] = a << b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    bo, bh = (bo % (np.dtype(dtype).itemsize * 8)).astype(dtype), (bh % (np.dtype(dtype).itemsize * 8)).astype(dtype)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_binary_rshift(dtype, shape):
    def fkt(a, b, c):
        c[:] = a >> b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    bo, bh = (bo % (np.dtype(dtype).itemsize * 8)).astype(dtype), (bh % (np.dtype(dtype).itemsize * 8)).astype(dtype)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_augmented_mod(dtype, shape):
    def fkt(a, c):
        c[:] %= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_augmented_lshifts(dtype, shape):
    def fkt(a, c):
        c[:] <<= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, [10]), random(dtype, [10])
    ao, ah = (ao % (np.dtype(dtype).itemsize * 8)).astype(dtype), (ah % (np.dtype(dtype).itemsize * 8)).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if issubclass(dtype, np.integer) or dtype == int], shapes[1:]))
def test_augmented_rshift(dtype, shape):
    def fkt(a, c):
        c[:] >>= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, [10]), random(dtype, [10])
    ao, ah = (ao % (np.dtype(dtype).itemsize * 8)).astype(dtype), (ah % (np.dtype(dtype).itemsize * 8)).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Tests for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import hope
import itertools
import pytest

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype", dtypes)
def test_blocks_1(dtype):
    def fkt(a, b, c, d):
        e = b[6:9]
        a[2:5] = e
        a[0] = 0
        a[0] = c[1]
        b[:] = d
    (ao, ah), (bo, bh), (c, _), (d, _) = random(dtype, [10]), random(dtype, [10]), random(dtype, [10]), random(dtype, [10])
    fkt(ao, bo, c, d), hope.jit(fkt)(ah, bh, c, d)
    assert check(ao, ah)
    assert check(bo, bh)
    fkt(ao, bo, c, d), hope.jit(fkt)(ah, bh, c, d)
    assert check(ao, ah)
    assert check(bo, bh)

def test_blocks_2():
    def fkt():
        i = 1
        i += 1
        i = 3
        return i
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert check(ro, rh)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_blocks_localvars(a, b, c):
    x = a
    c[:] = x



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

def test_opt_pow_array():
    def fkt(a, i):
        return a[i]
    hope.config.rangecheck = True
    hfkt = hope.jit(fkt)
    a = np.array([0, 1, 2])
    try:
        hfkt(a, 0)
        hfkt(a, 1)
        hfkt(a, 2)
    except:
        assert False
    try:
        hfkt(a, -1)
        assert False
    except: 
        pass
    try:
        hfkt(a, 3)
        assert False
    except: 
        pass
    hope.config.rangecheck = False



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the supported functionality with class methods.

Note: this should be used with precaution as unboxing member variables causes some overhead

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import hope

prefix = "hope_class"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def external(a, b):
    return a + b

class DemoClass(object):
    def __init__(self):
        self.a = 1
        
    @hope.jit
    def run(self, b):
        return self.a + b

    def run_external(self, b):
        return external(self.a, b)


def example():
    dc = DemoClass()
    #jit compile a method of a class
    r = dc.run(2)
    assert r == 3

    #call jitted function from a class
    r = dc.run_external(2)
    assert r == 3
    
    
if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_binary_mult(dtype, shape):
    def fkt(a, b, c):
        c[:] = a * b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    ao = np.copysign(np.sqrt(np.abs(ao.astype(np.float64))).astype(dtype), ao).astype(dtype)
    ah = np.copysign(np.sqrt(np.abs(ah.astype(np.float64))).astype(dtype), ah).astype(dtype)
    bo = np.copysign(np.sqrt(np.abs(bo.astype(np.float64))).astype(dtype), bo).astype(dtype)
    bh = np.copysign(np.sqrt(np.abs(bh.astype(np.float64))).astype(dtype), bh).astype(dtype)
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_augmented_mult(dtype, shape):
    def fkt(a, c):
        c[:] *= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 4.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 4.), ah).astype(dtype)
    co, ch = np.copysign(np.power(np.abs(co), 1. / 4.), co).astype(dtype), np.copysign(np.power(np.abs(ch), 1. / 4.), ch).astype(dtype)
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the supported NumPy functions

http://pythonhosted.org/hope/lang.html#numpy-mathematical-functions

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import numpy as np
import hope

prefix = "hope_numpy"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def numpy_math(x):
    # **Trigonometric functions**
    a = np.sin(x)     #   Trigonometric sine, element-wise.
    a = np.cos(x)     #   Cosine elementwise.
    a = np.tan(x)     #   Compute tangent element-wise.
    a = np.arcsin(x)     #Inverse sine, element-wise.
    a = np.arccos(x)     #Trigonometric inverse cosine, element-wise.
    a = np.arctan(x)     #Trigonometric inverse tangent, element-wise.
    
    # **Hyperbolic functions**
    a = np.sinh(x)     #  Hyperbolic sine, element-wise.
    a = np.cosh(x)     #  Hyperbolic cosine, element-wise.
    a = np.tanh(x)     #  Compute hyperbolic tangent element-wise.
    
    
    # **Miscellaneous**
    a = np.exp(x)     #   Calculate the exponential of all elements in the input array.
    a = np.sum(x)     #          Return the sum of array elements.
    a = np.sqrt(x)     #         Return the positive square-root of an array, element-wise.
    a = np.ceil(x)     #         Return the ceiling of the input, element-wise.
    a = np.floor(x)     #        Return the floor of the input, element-wise.
    a = np.trunc(x)     #        Return the truncated value of the input, element-wise.
    a = np.fabs(x)     #            Compute the absolute values element-wise
    a = np.pi     #              Returns the pi constant

@hope.jit
def numpy_interp(x, xp, fp):
    return np.interp(x, xp, fp)

def example():
    numpy_math(1.5)
    
    xp = np.array([0.0, 1.0])
    fp = np.array([1.0, 2.0])
    x = 0.5
    r = numpy_interp(x, xp, fp)
    assert r == 1.5

    x = np.array([0.25, 0.5, 0.75])
    r = numpy_interp(x, xp, fp)
    assert np.all(r == [1.25, 1.5, 1.75])
    
if __name__ == '__main__':
    example()


# Copyright (C) 2015 ETH Zurich, Institute for Astronomy

'''
Created on Feb 16, 2016

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import hope
import pytest
import numpy as np

from test.utilities import setup_module, setup_method, teardown_module
from hope.exceptions import UnsupportedFeatureException

def test_negative_idx_1d():
    def fkt(a): a[-1] = 1
    ah = np.zeros(10)
    hope.jit(fkt)(ah)
    assert np.all(ah[:-1] == 0)
    assert ah[-1] == 1


def test_negative_slice_1d_lower():
    def fkt(a): a[-1:] = 1
    ah = np.zeros(10)
    with pytest.raises(UnsupportedFeatureException):
        hope.jit(fkt)(ah)

def test_negative_slice_1d_upper():
    def fkt(a): a[:-1] = 1
    ah = np.zeros(10)
    hope.jit(fkt)(ah)
    assert np.all(ah[:-1] == 1)
    assert ah[-1] == 0

def test_negative_slice_1d_both():
    def fkt(a): a[1:-1] = 1
    ah = np.zeros(10)
    hope.jit(fkt)(ah)
    assert np.all(ah[1:-1] == 1)
    assert ah[-1] == 0
    assert ah[0] == 0
    
def test_negative_idx_2d():
    def fkt(a): a[-1, -1] = 1
    ao, ah = np.zeros((10,10)), np.zeros((10,10))
    fkt(ao)
    hope.jit(fkt)(ah)
    assert np.all(ao == ah)


def test_negative_slice_2d_lower():
    def fkt(a): a[-1:, -1:] = 1
    ah = np.zeros((10,10))
    with pytest.raises(UnsupportedFeatureException):
        hope.jit(fkt)(ah)

def test_negative_slice_2d_upper():
    def fkt(a): a[:-1, :-1] = 1
    ao, ah = np.zeros((10,10)), np.zeros((10,10))
    fkt(ao)
    hope.jit(fkt)(ah)
    assert np.all(ao == ah)

def test_negative_slice_2d_both():
    def fkt(a): a[1:-1, 1:-1] = 1
    ao, ah = np.zeros((10,10)), np.zeros((10,10))
    fkt(ao)
    hope.jit(fkt)(ah)
    assert np.all(ao == ah)




# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the mathematical optimization

http://pythonhosted.org/hope/optimization.html#optimization

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import numpy as np
import hope

prefix = "hope_optimize"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.optimize = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def simplify(d):
    r = np.sin(d)**2 + np.cos(d)**2
    return r

@hope.jit
def cse(d):
    r = 2*d + 4*d + 8*d
    return r

@hope.jit
def pow_replacement(d):
    r = d**2 + d**4
    return r

def example():
    # simplified to 1
    r = simplify(0.5)
    assert r == 1

    # reduced to 14 * d
    r = cse(0.5)
    assert r == 7

    # replaces pow and cse:
    # T1 = d * d
    # r = T1 * (1 + T1)
    r = pow_replacement(2)
    assert r == 20


if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_bool_(dtype):
    def fkt(a):
        return np.bool_(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_int_(dtype):
    def fkt(a):
        return np.int_(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_intc(dtype):
    def fkt(a):
        return np.intc(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_int8(dtype):
    def fkt(a):
        return np.int8(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_int16(dtype):
    def fkt(a):
        return np.int16(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_int32(dtype):
    def fkt(a):
        return np.int32(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_int64(dtype):
    def fkt(a):
        return np.int64(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_uint8(dtype):
    def fkt(a):
        return np.uint8(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_uint16(dtype):
    def fkt(a):
        return np.uint16(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_uint32(dtype):
    def fkt(a):
        return np.uint32(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_uint64(dtype):
    def fkt(a):
        return np.uint64(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_float_(dtype):
    def fkt(a):
        return np.float_(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_float32(dtype):
    def fkt(a):
        return np.float32(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)

@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if dtype != float])
def test_func_float64(dtype):
    def fkt(a):
        return np.float64(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert type(co) == type(ch)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Aug 26, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import pytest
import ast
import itertools
import hope
from hope import _tosource
from hope import _transformer

def dummy(par, *args, **kwargs):
    for _ in range(1):
        continue
    assert True, "msg"

@pytest.mark.infrastructure
class TestToSource(object):
    
    def test_tosource_mod(self):
        self.exec_test_on_mod(hope._tosource)
        
    def test_wrapper_mod(self):
        self.exec_test_on_mod(hope._wrapper)
        
    def test_simple(self):
        mod_ast = _transformer.get_fkt_ast(dummy)
        mod_source = _tosource.tosource(mod_ast)
#         with open("dummy.py", "w") as f:
#             f.write(mod_source)
        mod_ast2 = ast.parse(mod_source)
        assert compare_ast(mod_ast, mod_ast2.body[0])
        
    def exec_test_on_mod(self, mod):
        modpath = mod.__file__
        if modpath.endswith('.pyc'):
            modpath = modpath[:-1]
        
        with open(modpath, "r") as f:
            source = f.read()
        mod_ast = ast.parse(source)
        mod_source = _tosource.tosource(mod_ast)
        mod_source = mod_source.replace("'\\n'", "@@@@")
        mod_source = mod_source.replace("\\n", "\n")
        mod_source = mod_source.replace("@@@@", "'\\n'")
        
#         with open("dummy.py", "w") as f:
#             f.write(mod_source)
        mod_ast2 = ast.parse(mod_source)
        assert compare_ast(mod_ast, mod_ast2)

def compare_ast(node1, node2):
    if type(node1) is not type(node2):
        raise Exception(node1, node2)
        return False
    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in ('lineno', 'col_offset', 'ctx'):
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True
    elif isinstance(node1, list):
        return all(itertools.starmap(compare_ast, zip(node1, node2)))
    else:
        return node1 == node2


# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


from hope._ast import *


class Dumper(NodeVisitor):
    """
    Generates a human readable string representation of a HOPE AST.
    Can be used to dump an AST for debugging purpose. Is internally
    used to generate consistent variable names etc. 
    """

    def visit_Number(self, node):
        return "{0!r}.{1}".format(node.value, PY_TYPE_CHAR[node.dtype])

    def visit_NewVariable(self, node):
        return "{0}.new".format(node.name)

    def visit_Variable(self, node):
        ret = node.name
        if not node.dtype is None:
            ret += ".{0}".format(PY_TYPE_CHAR[node.dtype])
        if len(node.shape) > 0:
            segments = []
            for lower, upper in node.shape:
                segments.append("{0}:{1}".format("" if lower is None else "{0!s}".format(lower), self.visit(upper)))
            ret += "[{0}]".format(",".join(segments))
        return ret

    def visit_Object(self, node):
        return "{0}{1}@{2}".format(("{0}->".format(self.visit(node.parent)) if node.parent is not None else ""), node.name, node.classname)

    def visit_ObjectAttr(self, node):
        ret = "{0}->{1}".format(self.visit(node.parent), node.name)
        if not node.dtype is None:
            ret += ".{0}".format(PY_TYPE_CHAR[node.dtype])
        if len(node.shape) > 0:
            segments = []
            for lower, upper in node.shape:
                segments.append("{0}:{1}".format("" if lower is None else "{0!s}".format(lower), self.visit(upper)))
            ret += "[{0}]".format(",".join(segments))
        return ret

    def visit_Dimension(self, node):
        if isinstance(node.variable, ObjectAttr):
            parent = node.variable.parent
            trace = [node.variable.name]
            while not parent is None:
                trace.insert(0, parent.name)
                parent = parent.parent
            return "{0}@{1}".format(".".join(trace), node.dim)
        elif isinstance(node.variable, Variable):
            return "{0}@{1}".format(node.variable.name, node.dim)
        else:
            raise Exception("Unknown type {0}".format(node.variable.name))
        
    def visit_DimensionSlice(self, node):
        return "{0}+{1}".format(self.visit_Dimension(node), self.visit(node.slice))

    def visit_View(self, node):
        ret = node.variable.name
        if not node.variable.dtype is None:
            ret += ".{0}".format(PY_TYPE_CHAR[node.variable.dtype])
        shape = []
        for variable_extent, extent in zip(node.variable.shape, node.extents):
            if isinstance(extent, tuple):
                lower, upper = extent
                if lower is None: lower = variable_extent[0]
                if isinstance(lower, Number) and lower.value == 0: lower = None
                if upper is None: upper = variable_extent[1]
                shape.append("{0}:{1}".format("" if lower is None else self.visit(lower), self.visit(upper)))
            else:
                shape.append(self.visit(extent))
        ret += "[{0}]".format(", ".join(shape))
        return ret

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Assign(self, node):
        return "{0} = {1}".format(self.visit(node.target), self.visit(node.value))

    def visit_AugAssign(self, node):
        return "{0} {1} {2}".format(self.visit(node.target), node.op, self.visit(node.value))

    def visit_UnaryOp(self, node):
        return "{0}{1}".format(node.op, self.visit(node.operand))

    def visit_BinOp(self, node):
        return "({0} {1} {2})".format(self.visit(node.left), node.op, self.visit(node.right))

    def visit_BoolOp(self, node):
        return "({0})".format(" {0} ".format(node.op).join([self.visit(value) for value in node.values]))

    def visit_Compare(self, node):
        return "({0} {1} {2})".format(self.visit(node.left), node.op, self.visit(node.comparator))

    def visit_If(self, node):
        ret = "if {0} {{\n\t{1}\n}}".format(self.visit(node.test), "\n\t".join(self.visit(node.body).split("\n")))
        if not node.orelse is None:
            ret += " else {{\n\t{0}\n}}".format("\n\t".join(self.visit(node.orelse).split("\n")))
        return ret

    def visit_For(self, node):
        return "for {0} in ({1}:{2}) {{\n\t{3}\n}}".format(self.visit(node.iter), self.visit(node.lower), self.visit(node.upper), "\n\t".join(self.visit(node.body).split("\n")))

    def visit_While(self, node):
        return "while {0} {{\n\t{1}\n}}".format(self.visit(node.test), "\n\t".join(self.visit(node.body).split("\n")))

    def visit_Call(self, node):
        keywords = ["{0}={1}".format(name, self.visit(value)) for name, value in list(node.keywords.items())]
        return "{0}({1})".format(self.visit(node.name), ", ".join([self.visit(arg) for arg in node.args] + keywords))

    def visit_GlobalFunction(self, node):
        return node.name

    def visit_HopeAttr(self, node):
        return "hope.{0}".format(node.name)

    def visit_NumpyAttr(self, node):
        return "numpy.{0}".format(node.name)

    def visit_NumpyContraction(self, node):
        return "{0} = numpy.{1}({2})".format(self.visit(node.variable), node.op, self.visit(node.value))

    def visit_Allocate(self, node):
        return "new {0}".format(self.visit(node.variable))

    def visit_Return(self, node):
        return "return {0}".format(self.visit(node.value))

    def visit_Block(self, node):
        if len(node.shape):
            segments = ""
            for lower, upper in node.shape:
                segments += "({0}:{1})".format("" if lower is None else self.visit(lower), self.visit(upper))
            return "{0} {{\n\t{1}\n}}".format(segments, "\n\t".join("\n".join([self.visit(expr) for expr in node.body]).split("\n")))
        else:
            return "\n".join([self.visit(expr) for expr in node.body])

    def visit_Body(self, node):
        return "\n".join([self.visit(block) for block in node.blocks])

    def visit_FunctionDef(self, node):
        args = []
        for arg in node.signature:
            if isinstance(arg, Variable):
                args.append("{0}{1!s} {2}".format(NPY_TYPE[arg.dtype], "" if len(arg.shape) == 0 else "^{0}".format(len(arg.shape)), arg.name))
            elif isinstance(arg, Object):
                args.append("{0} {1}".format(arg.classname, arg.name))
            else:
                raise Exception("Unknown signature argument: {0!s}".format(self.visit(arg)))
        return "{0}({1})".format(node.name, ", ".join(args)) + "\n\t" + ("\n\t".join(self.visit(node.body).split("\n"))) + "\n"

    def visit_Module(self, node):
        code = ""
        for fktcls in list(node.functions.values()):
            for fkt in fktcls:
                code += self.visit(fkt)
        return code



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import os
import pickle

from hope import config


def serialize(obj, name):
    """
    Write a pickled representation of obj to a file named ``name`` inside ``hope.config.prefix``

    :param obj: arbitrary object to serialize
    :type obj: mixed
    :param name: name of the object
    :type name: str
    """

    if not os.path.exists(config.prefix):
        os.makedirs(config.prefix)

    with open(os.path.join(config.prefix, "{0}.pck".format(name)), "wb") as fp:
        pickle.dump(obj, fp)


def unserialize(name):
    """
    Read an object named ``name`` form ``hope.config.prefix``. If the file does not exits ``unserialize`` returns ``None``

    :param name: name of the object
    :type name: str
    :returns: mixed -- unserialized object
    """

    if not os.path.exists(os.path.join(config.prefix, "{0}.pck".format(name))):
        return None

    with open(os.path.join(config.prefix, "{0}.pck".format(name)), "rb") as fp:
        return pickle.load(fp)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test control structures for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

def test_ifelse_scalar():
    def fkt(a, b, c):
        if b > 0:
            c[:] = -a
        else:
            c[:] = a
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(np.int_, [10]), (1, 1), random(np.int_, [10])
    fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)
    (ao, ah), (bo, bh), (co, ch) = random(np.int_, [10]), (0, 0), random(np.int_, [10])
    fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)

# @pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, [[10]]))
# @make_test
# def test_for_array(a, b, c):
#     for aa in a:
#         c[0] = a

@pytest.mark.parametrize("dtype", dtypes)
def test_for_range_1(dtype):
    def fkt(a, b, c):
        for i in range(10):
            c[:, i] = a[i, :]
            c[i, 1] = a[0, 1] + b[i, 5]
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, [10, 10]), random(dtype, [10, 10]), random(dtype, [10, 10])
    ao, ah, bo, bh = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype), (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    ro, rh = fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.skipif("sys.version_info[0] > 2")
@pytest.mark.parametrize("dtype", dtypes)
def test_for_xrange_1(dtype):
    def fkt(a, b, c):
        for i in xrange(10):
            c[:, i] = a[i, :]
            c[i, 1] = a[0, 1] + b[i, 5]
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, [10, 10]), random(dtype, [10, 10]), random(dtype, [10, 10])
    ao, ah, bo, bh = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype), (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    ro, rh = fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype", dtypes)
def test_for_range_2(dtype):
    def fkt(a, b, c):
        for i in range(0, 10):
            c[:, i] = a[i, :]
            c[i, 1] = a[0, 1] + b[i, 5]
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, [10, 10]), random(dtype, [10, 10]), random(dtype, [10, 10])
    ao, ah, bo, bh = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype), (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    assert check(co, ch)
    ro, rh = fkt(ao, bo, co), hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype", dtypes)
def test_for_iteration_vars(dtype):
    """
    Tests if iteration vars can be used for slicing
    Derived from ufig.plugin.resample
    """
    def fkt(a, b, size_x):
        n = 2
        for dx in range(2 * n + 1):
            b[n + 1:(size_x - n - 1)] += a[dx + 1:(size_x + dx - 2 * n - 1)]
                
    hfkt = hope.jit(fkt)
    size_x = 10
    ao, ah = random(dtype, [size_x])
    ao = (ao / (2 * size_x + 1)).astype(dtype)
    ah = (ah / (2 * size_x + 1)).astype(dtype)
    
    bo, bh = np.zeros((size_x), dtype), np.zeros((size_x), dtype)
    
    fkt(ao, bo, size_x)
    hfkt(ah, bh, size_x)
    
    assert check(bo, bh)


def test_while():
    def fkt():
        i = 10
        j = 0
        while i > 0:
            j += 1
            i -= 1
        return j
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert check(ro, rh)

# TODO: make test for for/wile with array and for/else wihle/else



# Copyright (c) 2013 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import numpy as np


def exp(x):
    """
    Calculate the exponential of the input.

    :param x: Input data
    :type x: floating point number
    """
    return np.exp(x)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_opt_pow_array(dtype, shape):
    def fkt(a, c):
        c[:] = a ** 2.0 + a ** 4 + a ** 7
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 8.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 8.), ah).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype", dtypes)
def test_opt_pow_scalar(dtype):
    def fkt(a, c):
        c[0] = a ** 2 + a ** 4 + a ** 7.0
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [1])
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 8.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 8.), ah).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_opt_neg_pow_array(dtype, shape):
    def fkt(a, c):
        c[:] = a ** -2 + a ** -4.0 + a ** -7
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 8.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 8.), ah).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype", dtypes)
def test_opt_neg_pow_scalar(dtype):
    def fkt(a, c):
        c[0] = a ** -2 + a ** -4.0 + a ** -7
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [1])
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 8.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 8.), ah).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co), hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_opt_basic_array(dtype, shape):
    def fkt(a, c):
        c[:] = (a + a) * a - 1 / a
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 4.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 4.), ah).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype", dtypes)
def test_opt_basic_scalar(dtype):
    def fkt(a, c):
        c[0] = (a + a) * a - 1 / a
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [1])
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 4.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 4.), ah).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

@pytest.mark.parametrize("dtype", [float])
def test_opt_basic_scalar(dtype):
    def fkt(a, c):
        c[0] = a**-2
    hope.config.optimize = True
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [1])
    ao, ah = 1. / np.power(np.abs(ao), 1. / 2.).astype(dtype), 1. / np.power(np.abs(ah), 1. / 2.).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    hope.config.optimize = False

# TODO: add test for supported functions
# TODO: add test where resulting ast is really tested (pow)




# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Tests for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil, copy
hope.rangecheck = True

# TODO: differentiate between TravisCI (50min limit) and local + different python versions
# TODO: implement complex
# min_dtypes = [np.int8, np.uint64, np.float32, np.float64]
min_dtypes = [np.int8, np.uint64, np.float64]
# dtypes = [np.longlong, np.ulonglong, np.int8, np.int32, np.uint16, np.uint64, np.float32, np.float64, int, float]

if sys.version_info[0] == 3: #PY3 seems to be slower on TravisCI
    dtypes = [np.ulonglong, np.int8, np.uint64, np.float32, int, float]
else:
    dtypes = [np.ulonglong, np.int8, np.int32, np.uint64, np.float32, int, float]
#shapes = [[], [1200], [30, 13]]
shapes = [[], [3, 4]]

try:
    os.environ["HUDSON_URL"]
    JENKINS = True
except KeyError:
    JENKINS = False

ATOL = {}
ATOL[np.int8] = 1
ATOL[np.int16] = 1
ATOL[np.int32] = 1
ATOL[np.int64] = 1
ATOL[np.uint8] = 1
ATOL[np.uint16] = 1
ATOL[np.uint32] = 1
ATOL[np.uint64] = 1
ATOL[np.float32] = np.finfo(np.float32).resolution
ATOL[np.float64] = np.finfo(np.float64).resolution
ATOL[int] = ATOL[np.int_]
ATOL[float] = ATOL[np.float_]
if not np.bool_ in ATOL:
    ATOL[np.bool_] = 1
ATOL[bool] = ATOL[np.bool_]
if not np.longlong in ATOL:
    ATOL[np.longlong] = 1
if not np.ulonglong in ATOL:
    ATOL[np.ulonglong] = 1


def random(dtype, shape):
    if len(shape) == 0:
        if issubclass(dtype, np.inexact) or dtype == float:
            val = dtype(2 * (np.finfo(dtype).max * np.random.random() - np.finfo(dtype).max / 2))
        else:
            val = np.fromstring(np.random.bytes(np.dtype(dtype).itemsize), dtype=np.dtype(dtype))[0]
            if val == np.iinfo(dtype).min:
                val += 1;
    else:
        if issubclass(dtype, np.inexact) or dtype == float:
            val = 2 * (np.finfo(dtype).max * np.random.random(shape) - np.finfo(dtype).max / 2).astype(dtype)
        else:
            val = np.fromstring(np.random.bytes(np.prod(shape) * np.dtype(dtype).itemsize), dtype=np.dtype(dtype)).reshape(shape)
            val[val == np.iinfo(dtype).min] += 1
    return val, copy.deepcopy(val)


def check(a, b):
    atol = ATOL[a.dtype.type] if hasattr(a, "dtype") else ATOL[type(a)]
    aval = a.astype(np.float64) if hasattr(a, "dtype") else np.float64(a)
    btol = ATOL[b.dtype.type] if hasattr(b, "dtype") else ATOL[type(b)]
    bval = b.astype(np.float64) if hasattr(b, "dtype") else np.float64(b)
    return np.allclose(aval, bval, atol=min(atol, btol))


def make_test(fkt):
    def run(dtype, shape):
        hfkt = hope.jit(fkt)
        (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
        ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
        for a, b in [(ao, ah), (bo, bh), (co, ch), (ro, rh)]:
            if not a is None or not b is None:
                assert check(a, b)
        ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
        for a, b in [(ao, ah), (bo, bh), (co, ch), (ro, rh)]:
            if not a is None or not b is None:
                assert check(a, b)
    return run


def setup_module(module):
    np.seterr(all="raise")


def setup_method(self, method):
    np.random.seed(42)


def teardown_module(module):
    # delete all module in the lib directory to make sure the functions are rebuilt every time
    if os.path.isdir(hope.config.prefix):
        for filename in os.listdir(hope.config.prefix):
            path = os.path.join(hope.config.prefix, filename)
            if os.path.isfile(path) and os.path.splitext(path)[1] in [".pck", ".so", ".pyc", ".py"] and (filename.startswith("fkt_") or filename.startswith("test_")):
                os.remove(path)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_augmented_pow(dtype, shape):
    def fkt(a, c): 
        c[:] **= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(np.uint8, shape), random(dtype, shape)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    if np.count_nonzero(co == 0) > 0: co[co == 0] += 1
    if np.count_nonzero(ch == 0) > 0: ch[ch == 0] += 1
    co, ch = np.copysign(np.sqrt(np.abs(co)), co).astype(dtype), np.copysign(np.sqrt(np.abs(ch)), ch).astype(dtype)
    ao, ah = np.power(np.abs(ao).astype(np.float64), 1. / co.astype(np.float64)).astype(dtype), np.power(np.abs(ah).astype(np.float64), 1. / ch.astype(np.float64)).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)

# TODO: fix for np.ulonglong and uint64, std::power produce different results
@pytest.mark.parametrize("dtype,shape", itertools.product([dtype for dtype in dtypes if not dtype in [np.uint64, np.ulonglong]], shapes[1:]))
def test_binary_pow(dtype, shape):
    def fkt(a, c):
        c[:] = a ** 2
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.sqrt(np.abs(ao)), ao).astype(dtype), np.copysign(np.sqrt(np.abs(ah)), ah).astype(dtype)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtypea,dtypeb", itertools.product(min_dtypes, min_dtypes))
def test_cross_pow_dtype(dtypea, dtypeb):
    if JENKINS and dtypea == np.float32 and dtypeb == np.int8:
        pytest.skip("Fails on debian: dtypea={0!s}, dtypeb={1!s}".format(dtypea, dtypeb))
    def fkt(a, b, c):
        c[:] = a ** b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtypea, [100]), random(np.int8, []), random(dtypea, [100])
    if bo == 0 or bh == 0:
        bo, bh = dtypeb(1), dtypeb(1)
    else:
        bo, bh = np.abs(bo).astype(dtypeb), np.abs(bh).astype(dtypeb)
    ao, ah = np.power(np.abs(ao).astype(np.float64), 1. / bo).astype(dtypea), np.power(np.abs(ah).astype(np.float64), 1. / bh).astype(dtypea)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
@make_test
def test_return_arr(a, b, c): 
    return a

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
def test_return_arr_expr(dtype, shape):
    def fkt(a, c):
        return a ** -2 + a ** -4.0 + a ** -7
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 8.), ao).astype(dtype), np.copysign(np.power(np.abs(ah), 1. / 8.), ah).astype(dtype)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
@make_test
def test_return_arrcrt_zeros_dtype(a, b, c):
    d = np.zeros(3, dtype=np.int_)
    d[:] = 2
    d[0] = 1
    return d

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
@make_test
def test_return_arrcrt_ones_dtype(a, b, c):
    d = np.ones(3, dtype=np.int_)
    d[:] = 2
    d[0] = 1
    return d

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
@make_test
def test_return_arrcrt_zeros(a, b, c):
    d = np.zeros(3)
    d[:] = 2
    d[0] = 1
    return d

# TODO: fix for np.ulonglong, np.longlong and uint64
@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if not dtype in [np.uint64, np.longlong, np.ulonglong]])
def test_return_scalar(dtype):
    def fkt(a): return a
    ao, ah = random(dtype, [])
    ro, rh = fkt(ao), hope.jit(fkt)(ah)
    assert check(ro, rh)

# TODO: fix for np.ulonglong, np.longlong and uint64
@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if not dtype in [np.uint64, np.longlong, np.ulonglong]])
def test_return_arrayscalar(dtype):
    def fkt(a): return a[2]
    ao, ah = random(dtype, [10])
    ro, rh = fkt(ao), hope.jit(fkt)(ah)
    assert check(ro, rh)

# TODO: fix for np.ulonglong, np.longlong and uint64
@pytest.mark.parametrize("dtype", [dtype for dtype in dtypes if not dtype in [np.uint64, np.longlong, np.ulonglong]])
def test_return_arrayscalarcpy(dtype):
    def fkt(a): 
        b = a[2]
        return b
    ao, ah = random(dtype, [10])
    ro, rh = fkt(ao), hope.jit(fkt)(ah)
    assert check(ro, rh)

def test_return_const_int():
    def fkt(): return 3
    ro, rh = fkt(), hope.jit(fkt)()
    assert check(ro, rh)

def test_return_const_float():
    def fkt(): return 3.
    ro, rh = fkt(), hope.jit(fkt)()
    assert check(ro, rh)

def test_return_true():
    def fkt():
        return True
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert ro is rh

def test_return_false():
    def fkt():
        return False
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert ro is rh



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import ast
import sys
import inspect

from hope._dump import Dumper
from hope._ast import *
import hope._cache as cache
from hope.exceptions import UnsupportedFeatureException


class SymbolNamesCollector(NodeVisitor):
    """
    Traverses a HOPE AST (resp. subtree) to collect variable names (symbols) 
    in order to identify if variables are only used locally (in a loop/block) 
    or multiple times (body scope). 
    """
    
    def visit_Number(self, node):
        return []
    def visit_NewVariable(self, node):
        return [node.name]
    def visit_Variable(self, node):
        symbols = [node.name]
        for upper, lower in node.shape:
            symbols += [] if lower is None else self.visit(lower)
            symbols += [] if upper is None else self.visit(upper)
        return list(set(symbols))
    def visit_Object(self, node): 
        return []
    def visit_ObjectAttr(self, node):
        return []
    def visit_Dimension(self, node):
        if isinstance(node.variable, ObjectAttr):
            trace = node.variable.getTrace()
            return [".".join(trace)]
        
        return [node.variable.name]
    def visit_DimensionSlice(self, node):
        return self.visit_Dimension(node)
    def visit_View(self, node):
        symbols = self.visit(node.variable)
        for upper, lower in node.shape:
            symbols += [] if lower is None else self.visit(lower)
            symbols += [] if upper is None else self.visit(upper)
        return list(set(symbols))
    def visit_Expr(self, node):
        return self.visit(node.value)
    def visit_Assign(self, node):
        return list(set(self.visit(node.target) + self.visit(node.value)))
    def visit_Reference(self, node):
        return list(set(self.visit(node.target) + self.visit(node.value)))
    def visit_AugAssign(self, node):
        return list(set(self.visit(node.target) + self.visit(node.value)))
    def visit_UnaryOp(self, node):
        return self.visit(node.operand)
    def visit_BinOp(self, node):
        return list(set(self.visit(node.left) + self.visit(node.right)))
    def visit_BoolOp(self, node):
        return list(set([symbol for value in node.values for symbol in self.visit(value)]))
    def visit_Compare(self, node):
        return list(set(self.visit(node.left) + self.visit(node.comparator)))
    def visit_If(self, node):
        return list(set(self.visit(node.test) + self.visit(node.body) + (self.visit(node.orelse) if not node.orelse is None else [])))
    def visit_For(self, node):
        return list(set(self.visit(node.iter) + self.visit(node.body)))
    def visit_While(self, node):
        return list(set(self.visit(node.test) + self.visit(node.body)))
    def visit_Call(self, node):
        return list(set(self.visit(node.name) + [symbol for arg in node.args + [value for value in list(node.keywords.values())] for symbol in self.visit(arg)]))
    def visit_GlobalFunction(self, node):
        return []
    def visit_HopeAttr(self, node):
        return []
    def visit_NumpyAttr(self, node):
        return []
    def visit_NumpyContraction(self, node):
        return list(set(self.visit(node.variable) + self.visit(node.value)))
    def visit_Allocate(self, node):
        return self.visit(node.variable)
    def visit_Return(self, node):
        return self.visit(node.value)
    def visit_Block(self, node):
        return list(set([symbol for expr in node.body for symbol in self.visit(expr)]))
    def visit_Body(self, node):
        return list(set([symbol for block in node.blocks for symbol in self.visit(block)]))

class AllocateVisitor(NodeVisitor):
    """
    Visits the HOPE AST to identify variables, which needs to be allocated
    """
    
    def __init__(self, variables):
        self.variables = variables
        
    def generic_visit(self, node):
        pass
    
    def visit_If(self, node):
        self.visit(node.body)
        if node.orelse is not None:
            self.visit(node.orelse)
            
    def visit_For(self, node):
        scope = node.iter.scope
        node.iter.scope = "body"
        self.visit(node.body)
        node.iter.scope = scope
        
    def visit_While(self, node):
        self.visit(node.body)
        
    def visit_Block(self, node):
        return [self.visit(block) for block in node.body]
    
    def visit_Body(self, node):
        symbolcollector = SymbolNamesCollector()
        symboldict = {block: set(symbolcollector.visit(block)) for block in node.blocks}

        def check_blocks(start):
            for ind, block in enumerate(node.blocks):
                if start < ind:
                    for other in node.blocks[:ind] + node.blocks[(ind + 1):]:
                        for symbol in symboldict[block].intersection(symboldict[other]):
                            if self.variables[symbol].scope == "block":
                                self.variables[symbol].scope = "body"
                                self.variables[symbol].allocated = True
                                newBlock = Block(Allocate(self.variables[symbol]))
                                symboldict[newBlock] = set(symbolcollector.visit(newBlock))
                                node.blocks = node.blocks[:ind] + [newBlock] + node.blocks[ind:]
                                return ind - 1
            return None

        start = -1
        while True:
            start = check_blocks(start)
            if start is None:
                break

        return [self.visit(block) for block in node.blocks]

class IterableFunctionVisitor(ast.NodeVisitor):
    """
    Analyzes an AST to identify if a called function is 
    applied element-wise on an array (can be vectorized) if 
    the array has to be passed as argument.
    """
    
    def __init__(self, namespace, stack):
        self.namespace, self.stack = namespace, stack
        
    def generic_visit(self, node):
        raise NotImplementedError("Token not implemented: {0}".format(ast.dump(node)))
    def visit_Expr(self, node):
        return self.visit(node.value)
    def visit_Str(self, node):
        return False
    def visit_Assign(self, node):
        if len(node.targets) != 1:
            raise Exception("Only single target assignemants are supported")
        return self.visit(node.targets[0]) and self.visit(node.value)
    def visit_AugAssign(self, node):
        return self.visit(node.target) and self.visit(node.value)
    def visit_Subscript(self, node):
        return self.visit(node.value) and self.visit(node.slice)
    def visit_UnaryOp(self, node):
        return self.visit(node.operand)
    def visit_BinOp(self, node):
        return self.visit(node.left) and self.visit(node.right)
    def visit_BoolOp(self, node):
        return all([self.visit(value) for value in node.values])
    def visit_Compare(self, node):
        if len(node.ops) != 1 or len(node.comparators) != 1:
            raise Exception("More than one ops or more than one comparators are not supported")
        return self.visit(node.left) and self.visit(node.comparators[0])
    def visit_Num(self, node):
        return True
    def visit_Name(self, node):
        return True
    def visit_Slice(self, node):
        return node.lower if node.lower is None else self.visit(node.lower) and node.upper if node.upper is None else self.visit(node.upper)
    def visit_ExtSlice(self, node):
        return all([self.visit(dim) for dim in node.dims])
    def visit_Index(self, node):
        return True
    def visit_If(self, node):
        return (self.body_visit(node.orelse) if len(node.orelse) > 0 else True) and self.visit(node.test) and self.body_visit(node.body)
    def visit_For(self, node):
        return False
    def visit_While(self, node):
        return False
    def visit_Attribute(self, node):
        import hope
        if isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Name) and isinstance(node.value.ctx, ast.Load) \
                and node.value.id in self.namespace and self.namespace[node.value.id] is hope:
            return True
        elif isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Name) and isinstance(node.value.ctx, ast.Load) \
                and node.value.id in self.namespace and self.namespace[node.value.id] is np:
            return node.attr != "sum"
        else:
            return False

    def visit_Call(self, node):
        if (isinstance(node.func, ast.Name) and isinstance(node.func.ctx, ast.Load)):
            called_fkt = self.namespace[node.func.id]
            if (inspect.isbuiltin(called_fkt) or called_fkt.__name__ == "_hope_callback") and hasattr(cache, str(id(called_fkt))):
                called_fkt = getattr(cache, str(id(called_fkt)))
            elif not inspect.isfunction(called_fkt):
                raise Exception("Function '{0}' not a unbound, pure python function: ({1})".format(self.fkt.__name__, ast.dump(node)))
            
            if not node.func.id in self.namespace or not inspect.isfunction(called_fkt):
                raise Exception("Function '{0}' not accessible form global scope of function: ({1})".format(self.fkt.__name__, ast.dump(node)))
            
            called_fkt_ast = get_fkt_ast(called_fkt)
            
            if called_fkt in self.stack:
                return True
            else:
                return IterableFunctionVisitor(called_fkt.__globals__, self.stack + [called_fkt]).visit(called_fkt_ast)
        elif isinstance(node.func, ast.Attribute):
            return self.visit(node.func)
        else:
            return False

    def visit_Return(self, node):
        return self.visit(node.value)
    def visit_FunctionDef(self, node):
        return self.body_visit(node.body)
    def body_visit(self, stmts):
        for stmt in stmts:
            if not self.visit(stmt):
                return False
        return True

class ASTTransformer(ast.NodeVisitor):
    """
    A visitor implementation, traversing a Python AST of a function to create a HOPE AST.
    
    :param modtoken: The :py:class:`hope._ast.Module`, which should contain the function's HOPE AST.
    """
    
    def __init__(self, modtoken):
        self.modtoken, self.dumper, self.stack, self.exprs, self.next = modtoken, Dumper(), [], None, 0

    def merge_shapes(self, left, right):
        for leftslice, rightslice in zip(left.shape, right.shape):
            if leftslice is None or rightslice is None:
                raise Exception("Invalid shape!")
            
            leftslicestr = "{0}:{1}".format("" if leftslice[0] is None else self.dumper.visit(leftslice[0]), self.dumper.visit(leftslice[1]))
            rightslicestr = "{0}:{1}".format("" if rightslice[0] is None else self.dumper.visit(rightslice[0]), self.dumper.visit(rightslice[1]))
            self.token.shapes[leftslicestr], self.token.shapes[rightslicestr] = leftslice, rightslice
            if leftslicestr != rightslicestr:
                if not leftslicestr in self.merged and not rightslicestr in self.merged:
                    self.merged[leftslicestr] = len(self.merged)
                    self.merged[rightslicestr] = self.merged[leftslicestr]
                elif not leftslicestr in self.merged:
                    self.merged[leftslicestr] = self.merged[rightslicestr]
                elif not rightslicestr in self.merged:
                    self.merged[rightslicestr] = self.merged[leftslicestr]
                elif self.merged[leftslicestr] != self.merged[rightslicestr]:
                    for name, group in list(self.merged.items()):
                        if group == self.merged[rightslicestr]:
                            self.merged[name] = self.merged[leftslicestr]

    def dump_shape(self, shape):
        return "[{0}]".format(",".join(["{0}:{1}".format("" if seg[0] is None else self.dumper.visit(seg[0]), self.dumper.visit(seg[1])) for seg in shape]))

    def generic_visit(self, node):
        raise NotImplementedError("Token not implemented: {0}".format(ast.dump(node)))

    def visit_Expr(self, node):
        try:
            return Expr(self.visit(node.value))
        except Exception as ex:
            from hope._tosource import tosource
            ex.args = ((ex.args[0] if ex.args else "") + "\nin line " + tosource(node),) + ex.args[1:]
            raise

    def visit_Assign(self, node):
        try:
            if len(node.targets) != 1:
                raise UnsupportedFeatureException("Only single target assignments are supported")
            
            target, value = self.visit(node.targets[0]), self.visit(node.value)
            
            if isinstance(target, NewVariable):
                self.variables[target.name] = Variable(target.name, copy.deepcopy(value.shape), value.dtype)
                target = self.variables[target.name]
            elif isinstance(target, Variable) and len(target.shape) == 0: pass
            elif isinstance(target, ObjectAttr): pass
            elif not isinstance(target, View):
                raise Exception("Assignments are only allowed to views or variables")
            
            # TODO: should we check dtypes?
            if len(target.shape) > 0 and len(value.shape) == 0: pass
            elif len(target.shape) != len(value.shape):
                raise Exception("Invalid shapes: {0!s} {1!s}".format(self.dump_shape(target.shape), self.dump_shape(value.shape)))
            
            else:
                self.merge_shapes(target, value)
                

            if isinstance(value, ObjectAttr) and not isinstance(target, View): # reference to member
                name = ".".join(value.getTrace())
                self.variables[name] = Variable(name, shape=copy.deepcopy(value.shape), dtype=value.dtype, scope="body")
                if not isinstance(target, ObjectAttr): # avoid that target is allocated
                    self.variables[target.name].scope = "body"
                    self.variables[target.name].allocated = True
                
                return Reference(target, value)
                
            return Assign(target, value)
        
        except Exception as ex:
            from hope._tosource import tosource
            ex.args = ((ex.args[0] if ex.args else "") + "\nin line " + tosource(node),) + ex.args[1:]
            raise

    def visit_AugAssign(self, node):
        try:
            target, value = self.visit(node.target), self.visit(node.value)
            if len(value.shape) == 0: pass
            elif len(target.shape) != len(value.shape):
                raise Exception("Invalid shapes: {0!s} {1!s}".format(self.dump_shape(target.shape), self.dump_shape(value.shape)))
            
            else:
                self.merge_shapes(target, value)
            # TODO: should we check dtypes?
            return AugAssign(node.op, target, value)
        except Exception as ex:
            from hope._tosource import tosource
            ex.args = ((ex.args[0] if ex.args else "") + "\nin line " + tosource(node),) + ex.args[1:]
            raise

    def visit_Subscript(self, node):
        return View(self.visit(node.value), self.visit(node.slice))

    def visit_UnaryOp(self, node):
        op = type(node.op).__name__
        operand = self.visit(node.operand)
        if isinstance(operand, Number):
            opname, _ = UNARY_OPERATORS[op]
            return operand if opname == "+" else Number(-operand.value)
        return UnaryOp(op, operand)

    def visit_BinOp(self, node):
        left, right = self.visit(node.left), self.visit(node.right)
        if len(left.shape) == 0 or len(right.shape) == 0: pass
        elif len(left.shape) != len(right.shape):
            raise Exception("Invalid shapes: {0!s} {1!s}".format(self.dump_shape(left.shape), self.dump_shape(right.shape)))
        else:
            self.merge_shapes(left, right)
        return BinOp(type(node.op).__name__, left, right)

    def visit_BoolOp(self, node):
        values = [self.visit(value) for value in node.values]
        for value in values[1:]:
            if len(values[0].shape) != len(value.shape):
                raise Exception("Invalid shapes: {0!s} {1!s}".format(self.dump_shape(values[0].shape), self.dump_shape(value.shape)))
            else:
                self.merge_shapes(values[0], value)
        return BoolOp(type(node.op).__name__, values)

    def visit_Compare(self, node):
        if len(node.ops) != 1 or len(node.comparators) != 1:
            raise UnsupportedFeatureException("More than one ops or more than one comparators are not supported")
        
        left, comparator = self.visit(node.left), self.visit(node.comparators[0])
        if len(left.shape) == 0 or len(comparator.shape) == 0: pass
        elif len(left.shape) != len(comparator.shape):
            raise Exception("Invalid shapes: {0!s} {1!s}".format(self.dump_shape(left.shape), self.dump_shape(comparator.shape)))
        else:
            self.merge_shapes(left, comparator)
        return Compare(node.ops[0], left, comparator)

    def visit_Num(self, node):
        return Number(node.n)

    def visit_Name(self, node):
        if node.id == "True":
            return Number(True)
        elif node.id == "False":
            return Number(False)
        elif not node.id in self.variables:
            if isinstance(node.ctx, ast.Store):
                variable = NewVariable(node.id)
                self.variables[node.id] = variable
                return variable
            raise Exception("Unknown variable: {0}".format(node.id))
        
        return self.variables[node.id]

    def visit_NameConstant(self, node): #new in Py34
        return Number(node.value)

    def visit_Slice(self, node):
        if not node.step is None:
            raise Exception("Step size other than 1 are not supported")
        lower = node.lower if node.lower is None else self.visit(node.lower)
        upper = node.upper if node.upper is None else self.visit(node.upper)
        if isinstance(lower, Number) and lower.value < 0:
            raise UnsupportedFeatureException("Negative slices not supported")
#         if isinstance(upper, Number) and upper.value < 0:
#             raise UnsupportedFeatureException("Negative slices not supported")
        return [(lower, upper)]

    def visit_ExtSlice(self, node):
        return [slice for dim in node.dims for slice in self.visit(dim)]

    def visit_Index(self, node):
        if isinstance(node.value, ast.Tuple):
            return [self.visit(elt) for elt in node.value.elts]
        else:
            return [self.visit(node.value)]

    def visit_If(self, node):
        return If(self.visit(node.test), self.body_visit(node.body), self.body_visit(node.orelse) if len(node.orelse) > 0 else None)

    def visit_For(self, node):
        if len(node.orelse) > 0:
            raise UnsupportedFeatureException("else conditions in forloops are not supported")
        
        iter = self.visit(node.target)
        if not isinstance(iter, NewVariable):
            raise Exception("The variable '{0}' does already exists, since the scopeing is different in c++ and python, this is not supported".format(iter.name))
        # TODO: implement this more generic
        if isinstance(node.iter, ast.Call) and isinstance(node.iter.func, ast.Name) and node.iter.func.id in ["range", "xrange"] \
                and len(node.iter.keywords) == 0 and node.iter.starargs is None and node.iter.kwargs is None:
            if len(node.iter.args) == 1:
                args = [Number(0), self.visit(node.iter.args[0])]
            elif len(node.iter.args) == 2:
                args = [self.visit(arg) for arg in node.iter.args]
            else:
                raise UnsupportedFeatureException("only forloops with a call to range(lower, upper) or range(upper) are supported")
            
            self.variables[iter.name] = Variable(iter.name, [], np.int_)
            iter = self.variables[iter.name]
            return For(iter, args[0], args[1], self.body_visit(node.body))
        else:
            raise UnsupportedFeatureException("only forloops with a call to [x]range are supported")

    def visit_While(self, node):
        if len(node.orelse) > 0:
            raise UnsupportedFeatureException("else conditions in whileloops are not supported")
        
        return While(self.visit(node.test), self.body_visit(node.body))

    def visit_Attribute(self, node):
        import hope
        if isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Name) and isinstance(node.value.ctx, ast.Load) \
                and node.value.id in self.fkt.__globals__ and self.fkt.__globals__[node.value.id] is hope:
            return HopeAttr(node.attr)
        elif isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Name) and isinstance(node.value.ctx, ast.Load) \
                and node.value.id in self.fkt.__globals__ and self.fkt.__globals__[node.value.id] is np:
            if node.attr == "pi":
                return Number(np.pi)
            else:
                return NumpyAttr(node.attr)
        elif (isinstance(node.ctx, ast.Load) or isinstance(node.ctx, ast.Store)) and isinstance(node.value, ast.Name) and isinstance(node.value.ctx, ast.Load) \
                and node.value.id in self.variables and isinstance(self.variables[node.value.id], Object) \
                and (node.attr in self.variables[node.value.id].attrs or hasattr(self.variables[node.value.id].instance, node.attr)):
            return self.variables[node.value.id].getAttr(node.attr)
        elif isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Attribute):
            return self.visit(node.value).getAttr(node.attr)
        else:
            raise UnsupportedFeatureException("The attribute is not supported: {0}".format(ast.dump(node)))

    def visit_Call(self, node):
        #TODO: refactor
        # TODO: implement from bincount import bincount
        if (
                (isinstance(node.func, ast.Name) and isinstance(node.func.ctx, ast.Load))
            or (
                    isinstance(node.func, ast.Attribute) and node.func.value.id in self.variables and isinstance(self.variables[node.func.value.id], Object) \
                and (node.func.attr in self.variables[node.func.value.id].attrs or hasattr(self.variables[node.func.value.id].instance, node.func.attr)) \
                and inspect.ismethod(getattr(self.variables[node.func.value.id].instance, node.func.attr)) \
               )
        ):
            if isinstance(node.func, ast.Name) and node.func.id in self.fkt.__globals__:
                called_fkt = self.fkt.__globals__[node.func.id]
                if (inspect.isbuiltin(called_fkt) or called_fkt.__name__ == "_hope_callback") and hasattr(cache, str(id(called_fkt))):
                    called_fkt = getattr(cache, str(id(called_fkt)))
                elif not inspect.isfunction(called_fkt):
                    raise Exception("Function '{0}' not a unbound, pure python function: ({1})".format(self.fkt.__name__, ast.dump(node)))
                
                if not node.func.id in self.fkt.__globals__ or not inspect.isfunction(called_fkt):
                    raise Exception("Function '{0}' not accessible form global scope of function: ({1})".format(self.fkt.__name__, ast.dump(node)))
                
                args = [self.visit(arg) for arg in node.args]
                
            elif isinstance(node.func, ast.Attribute):
                called_fkt = getattr(self.variables[node.func.value.id].instance.__class__, node.func.attr)
                if (inspect.isbuiltin(called_fkt) or called_fkt.__name__ == "_hope_callback") and hasattr(cache, str(id(called_fkt))):
                    called_fkt = getattr(cache, str(id(called_fkt)))
                elif not inspect.isfunction(called_fkt) and not inspect.ismethod(called_fkt):
                    raise Exception("Function '{0}' not a unbound, pure python function: ({1})".format(self.fkt.__name__, ast.dump(node)))
                
                args = [self.variables[node.func.value.id]] + [self.visit(arg) for arg in node.args]
                
            else:
                raise Exception("Function '{0}' not accessible form global scope of function '{1}': ({2})".format(node.func.id, self.fkt.__name__, ast.dump(node)))
            
            if len(node.keywords) > 0:
                raise Exception("Keyword args are not allowed for global functions: {0} ({1})".format(self.fkt.__name__, ast.dump(node)))
            
            if not called_fkt.__defaults__ is None:
                raise Exception("Default args are not allowed for global functions: {0} ({1})".format(self.fkt.__name__, ast.dump(node)))
            
            if not called_fkt.__name__ in self.modtoken.functions:
                self.module_visit(called_fkt, args)
            valid = True
            if called_fkt is self.fkt:
                for arg, sig in zip(args, self.token.signature):
                    if isinstance(arg, Object) and arg.getId() != sig.getId():
                        valid = False
                    elif arg.dtype != sig.dtype or len(arg.shape) != len(sig.shape):
                        valid = False
                        
                if valid and self.token.shape is None:
                    raise Exception("Recursive call are only supported if a typed return occurs before the recursion in function: {0}".format(called_fkt.__name__))
                
                elif valid:
                    dtype, shape = self.token.dtype, self.token.shape
                    
            if not called_fkt is self.fkt or not valid:
                valid = False
                for fkt in self.modtoken.functions[called_fkt.__name__]:
                    if not valid:
                        valid = True
                        for arg, sig in zip(args, fkt.signature):
                            if isinstance(arg, Object) and arg.getId() != sig.getId():
                                valid = False
                            elif not isinstance(arg, Object) and (arg.dtype != sig.dtype or len(arg.shape) != len(sig.shape)):
                                valid = False
                                
                        if valid:
                            dtype, shape = fkt.dtype, fkt.shape
            if not valid:
                self.module_visit(called_fkt, args)
                for fkt in self.modtoken.functions[called_fkt.__name__]:
                    valid = True
                    for arg, sig in zip(args, fkt.signature):
                        if isinstance(arg, Object) and arg.getId() != sig.getId():
                            valid = False
                        elif not isinstance(arg, Object) and (arg.dtype != sig.dtype or len(arg.shape) != len(sig.shape)):
                            valid = False
                    if valid:
                        dtype, shape = fkt.dtype, fkt.shape
                        break
            if valid:
                return Call(GlobalFunction(called_fkt.__name__, [] if shape is None else shape, dtype), IterableFunctionVisitor(self.fkt.__globals__, []).visit(node), args)
            else:
                raise Exception("Unknown global function: {0}".format(node.func.id))

        elif isinstance(node.func, ast.Attribute) and isinstance(node.func.ctx, ast.Load) \
                and isinstance(node.func.value, ast.Name) and isinstance(node.func.value.ctx, ast.Load) \
                and node.func.value.id in self.fkt.__globals__ and self.fkt.__globals__[node.func.value.id] is np \
                and node.func.attr == "sum":
            
            name, self.next = "__sum{0}".format(self.next), self.next + 1
            args, keywords = [self.visit(arg) for arg in node.args], dict([(keyword.arg, self.visit(keyword.value)) for keyword in node.keywords])
            for keyword in list(keywords.keys()):
                # TODO: support axis, keepdims
                if keyword != "dtype":
                    raise UnsupportedFeatureException("HOPE only supports 'dtype' for numpy.sum at the moment. Received: %s"%keyword)
                
            if len(args) != 1:
                raise UnsupportedFeatureException("numpy.sum requires one argument. Received: %s"%(len(args)))
            
            self.variables[name] = Variable(name, [], getattr(np, keywords["dtype"].name) if "dtype" in keywords else args[0].dtype)
            self.exprs.append(NumpyContraction(self.variables[name], "sum", args[0]))
            return self.variables[name]
            
        elif isinstance(node.func, ast.Attribute):
            if not node.starargs is None or not node.kwargs is None:
                raise UnsupportedFeatureException("Only arguments without default values are supported in calls")
            
            return Call(self.visit(node.func), 
                        IterableFunctionVisitor(self.fkt.__globals__, []).visit(node), 
                        [self.visit(arg) for arg in node.args], 
                        dict([(keyword.arg, self.visit(keyword.value)) for keyword in node.keywords])
                        )

        else:
            raise Exception("Invalid call: {0}".format(ast.dump(node)))

    def visit_Return(self, node):
        value = self.visit(node.value)
        if len(value.shape) > 0 and not isinstance(value, (Variable, View)):
            name, self.next = "__ret{0}".format(self.next), self.next + 1
            self.variables[name] = Variable(name, copy.deepcopy(value.shape), value.dtype)
            self.merge_shapes(self.variables[name], value)
            self.exprs.append(Assign(self.variables[name], value))
            value = self.variables[name]
            
        if isinstance(value, (Variable, NewVariable)):
            # ensure Py_INCREF is added when appropriate to avoid mem leak or segfault
            self.token.return_allocated = value.allocated
        
        if not self.token.dtype is None and self.token.dtype != value.dtype:
            raise Exception("None unique return type")
        
        if not self.token.shape is None and (len(self.token.shape) == 0) != (len(value.shape) == 0):
            raise Exception("Return types needs to be either all scalar or all vectors")
        
        self.token.dtype, self.token.shape = value.dtype, value.shape
        return Return(value)

    def visit_FunctionDef(self, node):
        if len(node.args.defaults) or node.args.vararg is not None or node.args.kwarg is not None:
            raise Exception("Only positional arguments with no default values are supported")
        
        if len(self.args) != len(node.args.args):
            raise Exception("Function '%s' was called with invalid number of arguments. Expected: %s, received: %s"%(node.name, len(node.args.args), len(self.args)))

        signature = self._create_signature(node)

        self.token = FunctionDef(node.name, signature)

        # removing docstring
        if (isinstance(node.body[0], ast.Expr) and
            isinstance(node.body[0].value, ast.Str)):
            node.body.pop(0)
            
        # TODO: how do we parse recursive subfuncitons?
        self.token.body = self.body_visit(node.body) 
        AllocateVisitor(self.variables).visit(self.token.body)
        merged = {}
        for var, pos in list(self.merged.items()):
            merged[pos] = merged.get(pos, [])
            merged[pos].append(var)
        self.token.merged = list(merged.values())
        return self.token

    def body_visit(self, stmts):
        outerexprs, self.exprs, blocks, last_shape = self.exprs, [], [], "-1"
        for stmt in stmts:
            try:
                self.exprs.append(self.visit(stmt))            
            except Exception as ex:
                if "\nin body " not in (ex.args[0] if ex.args else ""):
                    from hope._tosource import tosource
                    ex.args = ((ex.args[0] if ex.args else "") + "\nin body " + tosource(stmt),) + ex.args[1:]
                raise
        for expr in self.exprs:
            segstrs = ["{0}:{1}".format("" if seg[0] is None else self.dumper.visit(seg[0]), self.dumper.visit(seg[1])) for seg in expr.shape]
            for segstr in segstrs:
                if not segstr in self.merged:
                    self.merged[segstr] = len(self.merged)
            shapestr = "/".join([repr(self.merged[segstr]) for segstr in segstrs])
            if len(shapestr) and len(blocks) > 0 and last_shape == shapestr:
                blocks[-1].body.append(expr)
            else:
                blocks.append(Block(expr))
            last_shape = shapestr
        self.exprs = outerexprs
        return Body(blocks)

    def module_visit(self, fkt, args, argid=None):
        """
        Converts the given function into a HOPE AST.
        First the source is extracted and parsed into a Python AST
        Then the AST is traversed with the arguments and a corresponding typed HOPE AST is built
        
        :param fkt: The function to be visited
        :param args: The argument, which have been passed to the function call
        :param argid: (optional) id made of the signature arguments
        """

        if argid is not None \
                and fkt.__name__ in self.modtoken.functions \
                and argid in [arg.getId() for arg in self.modtoken.functions[fkt.__name__]]:
            return

        if hasattr(self, "fkt"):
            self.stack.append((self.fkt, self.args, self.merged, self.variables, self.token))
        self.fkt, self.args, self.merged, self.variables, self.token = fkt, args, {}, {}, None

        fktast = get_fkt_ast(fkt)

        if len(fktast.args.defaults) or fktast.args.vararg is not None or fktast.args.kwarg is not None:
            raise Exception("Only positional arguments with no default values are supported")

        if not fkt.__name__ in self.modtoken.functions:
            self.modtoken.functions[fkt.__name__] = []

        if argid is None or not argid in self.modtoken.functions[fkt.__name__]:
            #Create the HOPE AST from the builtin AST
            parsed = self.visit(fktast)
            if not parsed.getId() in [arg.getId() for arg in self.modtoken.functions[fkt.__name__]]:
                self.modtoken.functions[fkt.__name__].append(parsed)

        if len(self.stack):
            self.fkt, self.args, self.merged, self.variables, self.token = self.stack.pop()
        else:
            delattr(self, "args"), delattr(self, "fkt")

    def _create_signature(self, node):
        """
        Creates a method signature for the given :py:class:`hope._ast.FunctionDef`
        
        :param node: instance of the FunctionDef
        
        :return signature: a list of :py:class:`hope._ast.Variable`
        """
        signature = []
        for name, arg in zip(node.args.args, self.args):
            if sys.version_info[0] == 2 and isinstance(name, ast.Name):
                if not isinstance(name.ctx, ast.Param):
                    raise Exception("Invalid Structure")
                
                argname = name.id
            elif sys.version_info[0] == 3 and isinstance(name, ast.arg):
                if not name.annotation is None:
                    raise Exception("Invalid Structure")
                
                argname = name.arg
            else:
                raise Exception("Invalid Structure")
            
            if isinstance(arg, (Variable, Object)):
                #TODO: maybe implement prototype pattern. 
                #Overwrite attributes of passed arguments to avoid conflicts
                self.variables[argname] = copy.copy(arg)
                self.variables[argname].name = argname
                self.variables[argname].scope = "signature"
                
            elif isinstance(arg, np.ndarray):
                if arg.dtype.isbuiltin != 1:
                    raise UnsupportedFeatureException("Only Builtin datatypes are supported: in {0} arg {1} has type {2!s}".format(self.fkt.__name__, argname, arg.dtype))
                
                self.variables[argname] = Variable(argname, None, arg.dtype.type, "signature", True)
                self.variables[argname].shape = [(None, Dimension(self.variables[argname], dim)) for dim in range(len(arg.shape))]
            elif hasattr(arg, "dtype"):
                if hasattr(arg.dtype, "type"):
                    self.variables[argname] = Variable(argname, [], arg.dtype.type, "signature", True)
                else:
                    self.variables[argname] = Variable(argname, [], arg.dtype, "signature", True)
            elif isinstance(arg, int):
                self.variables[argname] = Variable(argname, [], int, "signature", True)
            elif isinstance(arg, float):
                self.variables[argname] = Variable(argname, [], float, "signature", True)
            else:
                self.variables[argname] = Object(argname, arg)
            signature.append(self.variables[argname])
            
        return signature

def get_fkt_ast(fkt):
    """
    Creates a AST form the given function by getting the source of the function 
    by inspection, removing unneccesary indentations and passing the string 
    representation to the :py:mod:`ast` module. 
    
    :param fkt: Function object to be used
    
    :return fktast: the AST representation of the function
    
    :raises Exception: if the function could not be processed
    """
    lines = inspect.getsource(fkt).split("\n")
    for line in lines:
        if line.find("def") > -1 and (line.find("#") == -1 or line.find("#") > line.find("def")):
            outdent = line.find("def")
            break
    source = "\n".join([line[outdent:] for line in lines])
    modast = ast.parse(source)

    if not isinstance(modast, ast.Module) or len(modast.body) != 1 or not isinstance(modast.body[0], ast.FunctionDef):
        raise Exception("Error decompiling function")
    fktast = modast.body[0]
    return fktast



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import os
import sys
import hashlib
import inspect
import warnings

from hope._wrapper import Wrapper
import hope._cache as cache
from hope import config
from hope import serialization
from hope._wrapper import get_config_attrs
from hope._wrapper import get_fkt_hash

# TODO: add test for hope.pow
# TODO: add test and doc for additional numpy functions
# TODO: implement self, self.cost, self.fkt
# TODO: replace function in global namespace by c pointer to get the native speed on the first run
# TODO: make two versions for np.int_ and int
# TODO: make hope.interp with retuns a callable object with 2^n basepoints and c pendant
# TODO: if dtype of argument is alreadu correct, do not cast it in c code
# TODO: add test for subscrpto with all types to get the same semantics
# TODO: controll structures
# TODO: check for PyArray_ISNOTSWAPPED
# TODO: merge scalar blocks and search subexpressions in all lines
# TODO: asserts of sizes
# TODO: make class derived from tuple which can hold several parameters and be passed to hope functions
# TODO: use int PyArrayObject.flags to check if data has to be copied http://docs.scipy.org/doc/numpy/reference/c-api.types-and-structures.html#PyArrayObject
# TODO: use https://github.com/workhorsy/py-cpuinfo/blob/master/cpuinfo.py to detect features
# TODO: other code generator: http://documen.tician.de/codepy/jit.html#module-codepy.jit
# TODO: make ufunc decorator http://docs.scipy.org/doc/numpy/reference/ufuncs.html like numba http://numba.pydata.org/numba-doc/0.12.1/ufuncs.html
# TODO: optimize pow with interger powers to multipications
# TODO: make tests for _dump
# TODO: add constants/class constants to hope
# TODO: use sympy to simpify and Common Subexpression Detection
# 		http://docs.sympy.org/latest/modules/rewriting.html
# 		http://docs.sympy.org/latest/modules/core.html#module-sympy.core.sympify


def jit(fkt):
    """
    Compiles a function to native code and return the optimized function. The new function has the performance of a compiled function written in C.

    :param fkt: function to compile to c
    :type fkt: function
    :returns: function -- optimized function

    This function can either be used as decorator

    .. code-block:: python

        @jit
        def sum(x, y):
            return x + y

    or as a normal function

    .. code-block:: python

        def sum(x, y):
            return x + y
        sum_opt = jit(sum)
    """

    if config.hopeless:
        return fkt

    argspec = inspect.getargspec(fkt)
    if argspec.varargs is not None or argspec.keywords is not None:
        raise ValueError("Jitted functions should not have *args or **kwargs")

    hash = hashlib.sha224(inspect.getsource(fkt).encode('utf-8')).hexdigest()
    filename = "{0}_{1}".format(fkt.__name__, hash)

    if not os.path.exists(config.prefix):
        os.makedirs(config.prefix)

    if not config.prefix in sys.path:
        sys.path.append(os.path.abspath(config.prefix))

    wrapper = Wrapper(fkt, hash)

    try:
        state = serialization.unserialize(filename)
        if not state is None:
            _check_state(fkt, state)
        else:
            raise ImportError("No state found.")

        if sys.version_info[0] == 2:
            module = __import__(state["filename"], globals(), locals(), [], -1)
        else:
            import importlib
            module = importlib.import_module(state["filename"])

        if "bound" in state and state["bound"]:
            def _hope_callback(*args):
                return module.run(*args)
            setattr(cache, str(id(_hope_callback)), fkt)
            return _hope_callback
        else:
            module.set_create_signature(wrapper.create_signature)
            setattr(cache, str(id(module.run)), fkt)
            return module.run

    except LookupError as le:
        if config.verbose:
            warnings.warn("Recompiling... Reason: {0}".format(le))
        wrapper._recompile = True
        return wrapper.callback
    except ImportError as ie:
        return wrapper.callback

def _check_state(fkt, state):
    for name in get_config_attrs():
        if name not in state or state[name] != getattr(config, name):
            raise LookupError("State is inconsistent with config. Inconsistent state key: [{0}].".format(name))
        
    if "main" not in state or "called" not in state or state["main"] != fkt.__name__:
        raise LookupError("State is inconsistent")
    
    for name, value in list((state["called"] if "called" in state else {}).items()):
        if name not in fkt.__globals__:
            #TODO: FIX! state of globals depends on the order of function in module. If called function comes later in the code we raise the error
            raise LookupError("State is inconsistent. Called function '%s' cannot be found in %s's global scope."%(name, fkt.__name__))

        glob_fkt = fkt.__globals__[name]
        if isinstance(glob_fkt, Wrapper):
            if "filename" in state and get_fkt_hash(glob_fkt.fkt) != value:
                raise LookupError("State is inconsistent. Hash(sha224) has changed")
        elif inspect.isbuiltin(glob_fkt) and hasattr(cache, str(id(glob_fkt))):
            if "filename" in state and get_fkt_hash(getattr(cache, str(id(glob_fkt)))) != value:
                raise LookupError("State is inconsistent. Hash(sha224) has changed")
        elif inspect.isfunction(glob_fkt):
            if "filename" in state and get_fkt_hash(glob_fkt) != value:
                raise LookupError("State is inconsistent. Hash(sha224) of called function '%s' has changed"%name)
        elif "filename" in state:
            raise LookupError("State is inconsistent.")



# HOPE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HOPE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with HOPE.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division, absolute_import, unicode_literals


__all__ = ["jit", "config", "serialize", "unserialize"]
__author__ = "Lukas Gamper, Joel Akeret"
__email__ = "hope@phys.ethz.ch"
__version__ = "0.6.0"
__credits__ = "ETH Zurich, Institute for Astronomy"

from hope import config
from hope.jit import jit
from hope.serialization import serialize, unserialize
from hope.options import enableUnsaveMath, disableUnsaveMath

from hope.exp import exp

# save hope version in compiled files
config.version = __version__



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import numpy as np

NPY_TYPE = {}
NPY_TYPE[np.int8] = str("int8")
NPY_TYPE[np.int16] = str("int16")
NPY_TYPE[np.int32] = str("int32")
NPY_TYPE[np.int64] = str("int64")
NPY_TYPE[np.uint8] = str("uint8")
NPY_TYPE[np.uint16] = str("uint16")
NPY_TYPE[np.uint32] = str("uint32")
NPY_TYPE[np.uint64] = str("uint64")
NPY_TYPE[np.float32] = str("float32")
NPY_TYPE[np.float64] = str("float64")
NPY_TYPE[int] = NPY_TYPE[np.int_]
NPY_TYPE[float] = NPY_TYPE[np.float_]
if not np.bool_ in NPY_TYPE:
	NPY_TYPE[np.bool_] = str("bool_")
NPY_TYPE[bool] = NPY_TYPE[np.bool_]
if not np.longlong in NPY_TYPE:
	NPY_TYPE[np.longlong] = str("longlong")
if not np.ulonglong in NPY_TYPE:
	NPY_TYPE[np.ulonglong] = str("ulonglong")

PY_C_TYPE = {}
PY_C_TYPE[np.int8] = "npy_int8"
PY_C_TYPE[np.int16] = "npy_int16"
PY_C_TYPE[np.int32] = "npy_int32"
PY_C_TYPE[np.int64] = "npy_int64"
PY_C_TYPE[np.uint8] = "npy_uint8"
PY_C_TYPE[np.uint16] = "npy_uint16"
PY_C_TYPE[np.uint32] = "npy_uint32"
PY_C_TYPE[np.uint64] = "npy_uint64"
PY_C_TYPE[np.float32] = "npy_float"
PY_C_TYPE[np.float64] = "npy_double"
PY_C_TYPE[int] = PY_C_TYPE[np.int_]
PY_C_TYPE[float] = PY_C_TYPE[np.float_]
if not np.bool_ in PY_C_TYPE:
	PY_C_TYPE[np.bool_] = "npy_bool"
PY_C_TYPE[bool] = PY_C_TYPE[np.bool_]
if not np.longlong in PY_C_TYPE:
	PY_C_TYPE[np.longlong] = "npy_longlong"
if not np.ulonglong in PY_C_TYPE:
	PY_C_TYPE[np.ulonglong] = "npy_ulonglong"

PY_TYPE_CHAR = {}
PY_TYPE_CHAR[np.int8] = np.dtype(np.int8).char
PY_TYPE_CHAR[np.int16] = np.dtype(np.int16).char
PY_TYPE_CHAR[np.int32] = np.dtype(np.int32).char
PY_TYPE_CHAR[np.int64] = np.dtype(np.int64).char
PY_TYPE_CHAR[np.uint8] = np.dtype(np.uint8).char
PY_TYPE_CHAR[np.uint16] = np.dtype(np.uint16).char
PY_TYPE_CHAR[np.uint32] = np.dtype(np.uint32).char
PY_TYPE_CHAR[np.uint64] = np.dtype(np.uint64).char
PY_TYPE_CHAR[np.float32] = np.dtype(np.float32).char
PY_TYPE_CHAR[np.float64] = np.dtype(np.float64).char
PY_TYPE_CHAR[int] = "J"
PY_TYPE_CHAR[float] = "D"
if not np.bool_ in PY_TYPE_CHAR:
	PY_TYPE_CHAR[np.bool_] = "p"
PY_TYPE_CHAR[bool] = "P"
if not np.longlong in PY_TYPE_CHAR:
	PY_TYPE_CHAR[np.longlong] = np.dtype(np.longlong).char
if not np.ulonglong in PY_TYPE_CHAR:
	PY_TYPE_CHAR[np.ulonglong] = np.dtype(np.ulonglong).char

NPY_SCALAR_TAG = {}
NPY_SCALAR_TAG[np.int8] = "Int8"
NPY_SCALAR_TAG[np.int16] = "Int16"
NPY_SCALAR_TAG[np.int32] = "Int32"
NPY_SCALAR_TAG[np.int64] = "Int64"
NPY_SCALAR_TAG[np.uint8] = "UInt8"
NPY_SCALAR_TAG[np.uint16] = "UInt16"
NPY_SCALAR_TAG[np.uint32] = "UInt32"
NPY_SCALAR_TAG[np.uint64] = "UInt64"
NPY_SCALAR_TAG[np.float32] = "Float"
NPY_SCALAR_TAG[np.float64] = "Double"
if not np.bool_ in NPY_SCALAR_TAG:
	NPY_SCALAR_TAG[np.bool_] = "Bool"
if not np.longlong in NPY_SCALAR_TAG:
	NPY_SCALAR_TAG[np.longlong] = "LongLong"
if not np.ulonglong in NPY_SCALAR_TAG:
	NPY_SCALAR_TAG[np.ulonglong] = "ULongLong"

NPY_TYPEENUM = {}
NPY_TYPEENUM[np.int8] = "NPY_INT8"
NPY_TYPEENUM[np.int16] = "NPY_INT16"
NPY_TYPEENUM[np.int32] = "NPY_INT32"
NPY_TYPEENUM[np.int64] = "NPY_INT64"
NPY_TYPEENUM[np.uint8] = "NPY_UINT8"
NPY_TYPEENUM[np.uint16] = "NPY_UINT16"
NPY_TYPEENUM[np.uint32] = "NPY_UINT32"
NPY_TYPEENUM[np.uint64] = "NPY_UINT64"
NPY_TYPEENUM[np.float32] = "NPY_FLOAT32"
NPY_TYPEENUM[np.float64] = "NPY_FLOAT64"
NPY_TYPEENUM[int] = NPY_TYPEENUM[np.int_]
NPY_TYPEENUM[float] = NPY_TYPEENUM[np.float_]
if not np.bool_ in NPY_TYPEENUM:
	NPY_TYPEENUM[np.bool_] = "NPY_BOOL"
NPY_TYPEENUM[bool] = NPY_TYPEENUM[np.bool_]
if not np.longlong in NPY_TYPEENUM:
	NPY_TYPEENUM[np.longlong] = "NPY_LONGLONG"
if not np.ulonglong in NPY_TYPEENUM:
	NPY_TYPEENUM[np.ulonglong] = "NPY_ULONGLONG"

UNARY_OPERATORS = {}
UNARY_OPERATORS["UAdd"] = ("+", lambda a: type(+a(1)) )
UNARY_OPERATORS["USub"] = ("-", lambda a: type(-a(1)) )

BINARY_OPERATORS = {}
BINARY_OPERATORS["Add"] = ("+", lambda a, b: type(a(1) + b(1)) )
BINARY_OPERATORS["Sub"] = ("-", lambda a, b: type(a(1) - b(1)) )
BINARY_OPERATORS["Mult"] = ("*", lambda a, b: type(a(1) * b(1)) )
BINARY_OPERATORS["Div"] = ("/", lambda a, b: type(a(1) / b(1)) )
BINARY_OPERATORS["FloorDiv"] = ("//", lambda a, b: type(a(1) // b(1)) )
BINARY_OPERATORS["Pow"] = ("**", lambda a, b: type(a(1) ** b(1)) )
BINARY_OPERATORS["Mod"] = ("%", lambda a, b: type(a(1) % b(1)) )
BINARY_OPERATORS["LShift"] = ("<<", lambda a, b: type(a(1) << b(1)) )
BINARY_OPERATORS["RShift"] = (">>", lambda a, b: type(a(1) >> b(1)) )
BINARY_OPERATORS["BitOr"] = ("|", lambda a, b: type(a(1) | b(1)) )
BINARY_OPERATORS["BitXor"] = ("^", lambda a, b: type(a(1) ^ b(1)) )
BINARY_OPERATORS["BitAnd"] = ("&", lambda a, b: type(a(1) & b(1)) )

BOOL_OPERATORS = {}
BOOL_OPERATORS["And"] = "&&"
BOOL_OPERATORS["Or"] = "||"

AUGMENTED_ASSIGN = {}
AUGMENTED_ASSIGN["Add"] = "+="
AUGMENTED_ASSIGN["Sub"] = "-="
AUGMENTED_ASSIGN["Mult"] = "*="
AUGMENTED_ASSIGN["Div"] = "/="
AUGMENTED_ASSIGN["FloorDiv"] = "//="
AUGMENTED_ASSIGN["Pow"] = "**="
AUGMENTED_ASSIGN["Mod"] = "%="
AUGMENTED_ASSIGN["LShift"] = "<<="
AUGMENTED_ASSIGN["RShift"] = ">>="
AUGMENTED_ASSIGN["BitOr"] = "|="
AUGMENTED_ASSIGN["BitXor"] = "^="
AUGMENTED_ASSIGN["BitAnd"] = "&="

COMPARE_OPERATORS = {}
COMPARE_OPERATORS["Eq"] = ("==", lambda a, b: type(a(1) == b(1)) )
COMPARE_OPERATORS["NotEq"] = ("!=", lambda a, b: type(a(1) != b(1)) )
COMPARE_OPERATORS["Lt"] = ("<", lambda a, b: type(a(1) < b(1)) )
COMPARE_OPERATORS["LtE"] = ("<=", lambda a, b: type(a(1) <= b(1)) )
COMPARE_OPERATORS["Gt"] = (">", lambda a, b: type(a(1) > b(1)) )
COMPARE_OPERATORS["GtE"] = (">=", lambda a, b: type(a(1) >= b(1)) )

# http://docs.scipy.org/doc/numpy/reference/routines.math.html
NPY_UNARY_FUNCTIONS = {}
NPY_UNARY_FUNCTIONS["sin"] = "std::sin"
NPY_UNARY_FUNCTIONS["cos"] = "std::cos"
NPY_UNARY_FUNCTIONS["tan"] = "std::tan"
NPY_UNARY_FUNCTIONS["arcsin"] = "std::asin"
NPY_UNARY_FUNCTIONS["arccos"] = "std::acos"
NPY_UNARY_FUNCTIONS["arctan"] = "std::atan"
# NPY_UNARY_FUNCTIONS["degrees"] = "degrees"
# NPY_UNARY_FUNCTIONS["radians"] = "radians"
# NPY_UNARY_FUNCTIONS["deg2rad"] = "deg2rad"
# NPY_UNARY_FUNCTIONS["rad2deg"] = "rad2deg"
NPY_UNARY_FUNCTIONS["sinh"] = "std::sinh"
NPY_UNARY_FUNCTIONS["cosh"] = "std::cosh"
NPY_UNARY_FUNCTIONS["tanh"] = "std::tanh"
# NPY_UNARY_FUNCTIONS["arcsinh"] = "arcsinh"
# NPY_UNARY_FUNCTIONS["arccosh"] = "arccosh"
# NPY_UNARY_FUNCTIONS["arctanh"] = "arctanh"
NPY_UNARY_FUNCTIONS["floor"] = "std::floor"
NPY_UNARY_FUNCTIONS["ceil"] = "std::ceil"
NPY_UNARY_FUNCTIONS["trunc"] = "std::trunc"
NPY_UNARY_FUNCTIONS["exp"] = "std::exp"
# NPY_UNARY_FUNCTIONS["expm1"] = "expm1"
# NPY_UNARY_FUNCTIONS["exp2"] = "exp2"
NPY_UNARY_FUNCTIONS["log"] = "std::log"
# NPY_UNARY_FUNCTIONS["log10"] = "log10"
# NPY_UNARY_FUNCTIONS["log2"] = "log2"
# NPY_UNARY_FUNCTIONS["log1p"] = "log1p"
# NPY_UNARY_FUNCTIONS["iO"] = "iO"
# NPY_UNARY_FUNCTIONS["sinc"] = "sinc"
# NPY_UNARY_FUNCTIONS["signbit"] = "signbit"
# NPY_UNARY_FUNCTIONS["frexp"] = "frexp"
NPY_UNARY_FUNCTIONS["sqrt"] = "std::sqrt"
# NPY_UNARY_FUNCTIONS["square"] = "square"
NPY_UNARY_FUNCTIONS["abs"] = "abs"
# NPY_UNARY_FUNCTIONS["absolute"] = "absolute"
NPY_UNARY_FUNCTIONS["fabs"] = "std::fabs"
NPY_UNARY_FUNCTIONS["sign"] = "sign"

NPY_BINARY_FUNCTIONS = {}
# NPY_BINARY_FUNCTIONS["hypot"] = "hypot"
# NPY_BINARY_FUNCTIONS["arctan2"] = "arctan2"
# NPY_BINARY_FUNCTIONS["logaddexp"] = "logaddexp"
# NPY_BINARY_FUNCTIONS["logaddexp2"] = "logaddexp2"
# NPY_BINARY_FUNCTIONS["copysign"] = "copysign"
# NPY_BINARY_FUNCTIONS["ldexp"] = "ldexp"
# NPY_BINARY_FUNCTIONS["power"] = "power"
# NPY_BINARY_FUNCTIONS["convolve"] = "convolve"
# NPY_BINARY_FUNCTIONS["maximum"] = "maximum"
# NPY_BINARY_FUNCTIONS["minimum"] = "minimum"
# NPY_BINARY_FUNCTIONS["fmax"] = "fmax"
# NPY_BINARY_FUNCTIONS["fmin"] = "fmin"

# Numpy type conversions
NPY_CAST_FUNCTIONS = {}
NPY_CAST_FUNCTIONS["bool_"] = np.bool_
NPY_CAST_FUNCTIONS["int_"] = np.int_
NPY_CAST_FUNCTIONS["intc"] = np.intc
NPY_CAST_FUNCTIONS["int8"] = np.int8
NPY_CAST_FUNCTIONS["int16"] = np.int16
NPY_CAST_FUNCTIONS["int32"] = np.int32
NPY_CAST_FUNCTIONS["int64"] = np.int64
NPY_CAST_FUNCTIONS["uint8"] = np.uint8
NPY_CAST_FUNCTIONS["uint16"] = np.uint16
NPY_CAST_FUNCTIONS["uint32"] = np.uint32
NPY_CAST_FUNCTIONS["uint64"] = np.uint64
NPY_CAST_FUNCTIONS["float_"] = np.float_
NPY_CAST_FUNCTIONS["float32"] = np.float32
NPY_CAST_FUNCTIONS["float64"] = np.float64



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_binary_div(dtype, shape):
    def fkt(a, b, c):
        c[:] = a / b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    if np.count_nonzero(bo == 0) > 0: bo[bo == 0] += 1
    if np.count_nonzero(bh == 0) > 0: bh[bh == 0] += 1
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    if dtype in [np.float32, np.float64, float]:
        co[co < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
        ch[ch < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_binary_floordiv(dtype, shape):
    def fkt(a, b, c):
        c[:] = a // b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    if np.count_nonzero(bo == 0) > 0: bo[bo == 0] += 1
    if np.count_nonzero(bh == 0) > 0: bh[bh == 0] += 1
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    if dtype in [np.float32, np.float64, float]:
        co[co < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
        ch[ch < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
    assert check(co, ch)

# TODO: fix for np.int8 / np.int8 
@pytest.mark.parametrize("dtypea,dtypeb,dtypec", itertools.product(min_dtypes, min_dtypes, min_dtypes))
def test_cross_div(dtypea, dtypeb, dtypec):
    if dtypea == np.int8 and dtypeb == np.int8:
        pytest.skip("Different behaviour in c++ and python for int8 / int8".format(dtypea, dtypeb))
    def fkt(a, b, c):
        c[:] = a / b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtypea, [10]), random(dtypeb, [10]), random(dtypec, [10])
    ao, ah, bo, bh = ao.astype(np.float64), ah.astype(np.float64), bo.astype(np.float64), bh.astype(np.float64)
    ao, ah = np.copysign(np.power(np.abs(ao), 1. / 4.), ao).astype(dtypea), np.copysign(np.power(np.abs(ah), 1. / 4.), ah).astype(dtypea)
    bo, bh = np.copysign(np.power(np.abs(bo), 1. / 4.), bo).astype(dtypeb), np.copysign(np.power(np.abs(bh), 1. / 4.), bh).astype(dtypeb)
    if np.count_nonzero(bo == 0) > 0: bo[bo == 0] += 1
    if np.count_nonzero(bh == 0) > 0: bh[bh == 0] += 1
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_augmented_div(dtype, shape):
    def fkt(a, c):
        c[:] /= a
    hfkt = hope.jit(fkt)
    
    (ao, ah) = random(dtype, shape)
    (co, ch) = random(np.float64, shape)
    
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    if dtype in [np.float32, np.float64, float]:
        co[co < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
        ch[ch < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
    assert check(co.astype(dtype), ch.astype(dtype))

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_augmented_floordiv(dtype, shape):
    def fkt(a, c):
        c[:] //= a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    if np.count_nonzero(ao == 0) > 0: ao[ao == 0] += 1
    if np.count_nonzero(ah == 0) > 0: ah[ah == 0] += 1
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    if dtype in [np.float32, np.float64, float]:
        co[co < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
        ch[ch < 1. /  (np.finfo(dtype).max * np.finfo(dtype).resolution)] /= np.finfo(dtype).resolution
    assert check(co.astype(dtype), ch.astype(dtype))



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from test.utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_lt(a, b, c): c[:] = a < b

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_gt(a, b, c): c[:] = a > b

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_le(a, b, c): c[:] = a <= b

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_ge(a, b, c): c[:] = a >= b

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_eq(a, b, c): c[:] = a == b

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_neq(a, b, c): c[:] = a != b

def test_true():
    def fkt():
        if True:
            return 1
        else:
            return 0
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert check(ro, rh)

def test_false():
    def fkt():
        if False:
            return 1
        else:
            return 0
    hfkt = hope.jit(fkt)
    ro, rh = fkt(), hfkt()
    assert check(ro, rh)



# Copyright (c) 2013 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


from hope.options import get_cxxflags

# Additional compiler flags, formated as array of strings
cxxflags = get_cxxflags()
""" 
List of c++ compiler flags. Normally hope does determing the right flags itself.
"""

#TODO implement
prefix = ".hope"
""" 
Prefix of the folder hope saves all data in.
"""

verbose = False
""" 
Print a intermediate representation of each function during compilation.
"""

optimize = False
"""
Use '''sympy''' to simplify expression and exptract common subexpression detection
"""

keeptemp = False
""" 
Keep the intermediate c++ source and compiler output generated during compilation.
"""

rangecheck = False
""" 
Check if indeces are out of bounds
"""

hopeless = False
""" 
Disable hope. If hope.config.hopeless is True, hope.jit return the original function.
Use this function for debug purpos
"""

# make readable cpp file, but typecasting is not exactly the same as in numpy - this flag is private
_readablecxx = False
